
------------------------------------------------------------------------
--
-- AbsorbsMonitor
--
-- Copyright (C) 2010  Philipp Schmidt
--
-- This program is free software; you can redistribute it and/or
-- modify it under the terms of the GNU General Public License
-- as published by the Free Software Foundation; either version 2
-- of the License, or (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to:
--
-- Free Software Foundation, Inc.,
-- 51 Franklin Street, Fifth Floor,
-- Boston, MA  02110-1301, USA.
--
--
------------------------------------------------------------------------


local AM_Public = LibStub("AceAddon-3.0"):NewAddon("AbsorbsMonitor");

local AM_Core = setmetatable({}, {
	__tostring = function()
		return "AM Debug";
	end
});

LibStub("AceConsole-3.0"):Embed(AM_Core);
LibStub("AceEvent-3.0"):Embed(AM_Core);
LibStub("AceTimer-3.0"):Embed(AM_Core);
LibStub("AceComm-3.0"):Embed(AM_Core);
LibStub("AceSerializer-3.0"):Embed(AM_Core);

AbsorbsMonitor = AM_Public;


-- #################################################################################################


local L = LibStub("AceLocale-3.0"):GetLocale("AbsorbsMonitor");

local playerGUID;
local playerClass;

-- Specifies the channel any AddOn message should be
-- sent to, nil if silent
-- Can also be used to get the last known party state
local curChatChannel = nil;

-- Always hold the timestamp from the last COMBAT_LOG_EVENT_UNFILTERED fired
local lastCombatLogEvent = 0.0;

-- Table of all active absorb effects indexed by GUID and then spellId
-- at spellId == -1 there is a numeric entry with the total remaining value,
-- at spellId == -2 there is a numeric entry with the total quality (that is, the minimal quality)
-- The priority is also in this table for performance reasons during sort
-- [GUID] = { [spellId] = {spellId, priority, remainingValue, maxValue, quality, durationTimerHandle, extra} }
local activeEffectsBySpell = {};

-- Table of all active absorb effects indexed by GUID and then a list in the order in which
-- they will be used
local activeEffectsByPriority = {};

-- Table of all active area absorb effects indexed by triggerGUID
-- same format as an entry in activeEffectsBySpell with two extra fields at the end
-- {..., triggerGUID, refcount}
-- Note that in this approach, a trigger can have only one absorb effect up at the same time.
-- This should be a reasonable assumption for quite some time, considering only Anti-Magic Zone
-- and (soon) Power Word: Barrier use this feature anyway
local activeAreaEffects = {};

-- Table of current unit charges
-- A charge is a variant value a custom trigger can put on any unit with a limited lifetime
-- It is organized as a (very simple) queue (FIFO) for use with Divine Aegis e.g. to save the
-- critical heal value and then apply it on the aura gain or with Val'anyr.
-- [GUID] = { [spellId] = { charge1, charge2, ... } }
local activeCharges = {};

-- Table of known spells that cause an absorb effects
-- priority: active effects above 2 are neither used in total value nor displayed (e.g. Anti-Magic Shell)
-- [spellId] = {priority, duration, createFunc, hitFunc}
-- actually AbsorbsMonitor.Effects, defined in Effects.lua
local AbsorbEffects;

-- Table of spells that cause an area effect
-- [spellId (trigger)] = spellId (absorb effect)
-- actually AbsorbsMonitor.AreaTriggers, defined in Effects.lua
local AreaTriggers;

-- Table of additional callbacks on combat log events for proc-based and other non-generic absorb effects.
-- actually AbsorbsMonitor.CombatTriggers, defined in Effects.lua
local CombatTriggersOnHeal;
local CombatTriggersOnHealCrit;
local CombatTriggersOnAuraApplied;
local CombatTriggersOnAuraRemoved;

-- Table of all unit stats relevant to absorb effects like attack power and spell power
-- (mastery rating later on)
-- [GUID] = { class, AttackPower, SpellPower, quality }
local UnitStats;

-- Table of all scaling factors to absorb effects like talents, items, set boni, buffs
-- If there is no mechanism in Cataclysm to obtain the correct absorb amount by any effect
-- this entries are meant to be distributed among a group, raid, etc
-- [GUID] = { [scaling_Name] = scaling_Value }
-- A numerical scaling_name above 10 should ONLY be used if it is the spellId of the affected effect,
-- since it is sometimes used as a very quick way to check for a unit's class
-- Scaling factors that are only relevant to the local user like priest talents and do not
-- need to be distributed but are needed for calculation of the public one's are private ones,
-- found at index -1
local Scaling;

-- Class-specific callbacks
AM_Public.OnEnableClass = {};
AM_Public.OnScalingDecode = {};


-- #################################################################################################


local CommStatsCooldown = false;

local function ClearCommStatsCooldown()
	CommStatsCooldown = false;
end

local CommScalingCooldown = false;

local function ClearCommScalingCooldown()
	CommScalingCooldown = false;
end

local function DummyFunc()
end

local function DeepTableCopy(src)
	dest = {};
	
	for k, v in pairs(src) do
		if(type(k) == "table") then
			k = DeepTableCopy(k);
		end
		
		if(type(v) == "table") then
			v = DeepTableCopy(v);
		end
		
		dest[k] = v;
	end
	
	setmetatable(dest, getmetatable(src));
	
	return dest;
end

local function SortEffects(a, b)
	if(a[2] == b[2]) then
		return (a[3] < b[3]);
	else
		return (a[2] > b[2]);
	end
end

-- Only meant for ACTIVE (BUFF, AREA) effects
-- Note that this method should NOT be called on non-existing effects or units
-- There are no exist checks within it.
local function RemoveActiveEffect(guid, spellId)
	local guidEffects = activeEffectsBySpell[guid];
	local effectEntry = guidEffects[spellId];
	
	--AM_Core:Print("Removing absorb effect "..spellId.." on "..guid.." with "..guidEffects[spellId][3].." left, total: "..guidEffects[-1].." ("..#(activeEffectsByPriority[guid])..")");
			
	-- This is a shared effect with a triggerGUID
	if(effectEntry[8]) then
		effectEntry[9] = effectEntry[9] - 1;
		
		if(guid ~= effectEntry[8]) then
			AM_Core:SendMessage("Absorbs_EffectRemoved", guid, spellId);
		end
		
		if(effectEntry[9] == 0) then
			if(effectEntry[6]) then
				AM_Core:CancelTimer(effectEntry[6], true);
			end
			
			AM_Core:SendMessage("Absorbs_AreaCleared", effectEntry[8]);
		end	
	
	else
		if(effectEntry[2] < 2) then
			guidEffects[-1] = guidEffects[-1] - effectEntry[3];		
			
			AM_Core:SendMessage("Absorbs_UnitUpdated", guid, guidEffects[-1], guidEffects[-2]);
		end
	
		AM_Core:CancelTimer(effectEntry[6], true);
		
		AM_Core:SendMessage("Absorbs_EffectRemoved", guid, spellId);	
	end		
	
	if(#(activeEffectsByPriority[guid]) == 1) then		
		activeEffectsBySpell[guid] = nil;
		activeEffectsByPriority[guid] = nil;
				
		AM_Core:SendMessage("Absorbs_UnitCleared", guid);
	else
		guidEffects[spellId] = nil;
		
		for k, v in pairs(activeEffectsByPriority[guid]) do
			if(v[1] == spellId) then
				tremove(activeEffectsByPriority[guid], k);
				
				break;
			end
		end
	end
end

local function OnSingularTimeout(args)
	local guid, spellId = args[1], args[2];

	if(activeEffectsBySpell[guid] and activeEffectsBySpell[guid][spellId]) then
		RemoveActiveEffect(guid, spellId);
	end
end

local function OnSingularActivityCheck(args)
	local guid, spellId = args[1], args[2];
	
	if(activeEffectsBySpell[guid] and activeEffectsBySpell[guid][spellId]) then
		local _, _, _, _, _, name = GetPlayerInfoByGUID(guid);
		
		-- We cannot track whether it's still on, remove it
		if(not name) then
			AM_Core:CancelTimer(activeEffectsBySpell[guid][spellId][6]);
		
			RemoveActiveEffect(guid, spellId);
		end
		
		local stillActive = false;
		
		for i=1, 40 do
			local _, _, _, _, _, _, _, _, _, _, buffId = UnitBuff(name, i);
			
			if(not buffId) then break; end
			
			if(buffId == spellId) then
				stillActive = true;
				
				break;
			end
		end
		
		if(not stillActive) then
			AM_Core:CancelTimer(activeEffectsBySpell[guid][spellId][6]);
		
			RemoveActiveEffect(guid, spellId);
		end
	else
		-- Make sure to remove the timer
		AM_Core:CancelTimer(activeEffectsBySpell[guid][spellId][6]);
	end
end

-- This should be called in even less cases than the normal timeout
local function OnAreaTimeout(areaEntry)
	-- Disable the timer handle entry already
	areaEntry[6] = nil;

	for guid, guidEffects in pairs(activeEffectsBySpell) do
		for spellId, effectEntry in pairs(guidEffects) do
			if(effectEntry == areaEntry) then				
				RemoveActiveEffect(guid, spellId);
				
				if(areaEntry[9] == 0) then return; end
			end
		end
	end
	
	-- We're only here if we didn't reduce the refcount to zero
	error("Positive refcount "..areaEntry[9].." remained for area effect "..areaEntry[1].." by trigger "..areaEntry[8]);
end

local function ApplySingularEffect(sourceGUID, sourceName, destGUID, destName, spellId)
	local destEffects = activeEffectsBySpell[destGUID];
	local effectInfo = AbsorbEffects[spellId];
	local effectEntry;
	
	local value, quality, extra = effectInfo[3](sourceGUID, sourceName, destGUID, destName, spellId, destEffects);
	
	if(value == nil) then return; end
	
	-- No entry yet for this unit
	if(not destEffects) then
		effectEntry = {spellId, effectInfo[1], value, value, quality, 0, extra};
		
		destEffects = {[-1] = 0, [-2] = 1.0, [spellId] = effectEntry};
		
		activeEffectsBySpell[destGUID] = destEffects
		activeEffectsByPriority[destGUID] = {effectEntry};
		
		AM_Core:SendMessage("Absorbs_EffectApplied", sourceGUID, sourceName, destGUID, destName, spellId, value, quality, effectInfo[2]);
	
	-- Not this specific effect yet
	elseif(not destEffects[spellId]) then	
		effectEntry = {spellId, effectInfo[1], value, value, quality, 0, extra};
				
		destEffects[spellId] = effectEntry;
	
		tinsert(activeEffectsByPriority[destGUID], effectEntry);
		sort(activeEffectsByPriority[destGUID], SortEffects);
		
		AM_Core:SendMessage("Absorbs_EffectApplied", sourceGUID, sourceName, destGUID, destName, spellId, value, quality, effectInfo[2]);
	
	-- Effect exists already
	else	
		effectEntry = destEffects[spellId];
		local prevAmount = effectEntry[3];
	
		effectEntry[3] = value;
		effectEntry[4] = value;
		effectEntry[5] = quality;
		effectEntry[7] = extra;
	
		AM_Core:SendMessage("Absorbs_EffectUpdated", destGUID, spellId, value, quality, effectInfo[2]);
	
		-- Adjust value in case this is a visible absorb to get the difference
		value = value - prevAmount;
		
		-- Cancel the exting duration timeout timer
		AM_Core:CancelTimer(effectEntry[6], true);
	end
	
	if(quality < destEffects[-2]) then
		destEffects[-2] = quality;
	end
	
	if(effectInfo[1] < 2) then
		destEffects[-1] = destEffects[-1] + value;
	
		AM_Core:SendMessage("Absorbs_UnitUpdated", destGUID, destEffects[-1], destEffects[-2]);
	end

	
	if(effectInfo[2]) then
		-- We add a 5s grace period for latency and all kind of stuff
		-- This duration timeout should only be needed if the unit moved out of combat log
		-- reporting range anyway
		effectEntry[6] = AM_Core:ScheduleTimer(OnSingularTimeout, effectInfo[2] + 5, {destGUID, spellId});
	else
		effectEntry[6] = AM_Core:ScheduleRepeatingTimer(OnSingularActivityCheck, 8, {destGUID, spellId});
	end
	
	--AM_Core:Print("Applied buff effect "..spellId.." (priority: "..effectInfo[1]..") on "..destGUID.." for "..value..", new total: "..activeEffectsBySpell[destGUID][0].." ("..#(activeEffectsByPriority[destGUID])..")");
end

local function ApplyAreaEffect(triggerGUID, triggerName, destGUID, destName, spellId)	
	if(not activeAreaEffects[triggerGUID]) then
		return ApplySingularEffect(triggerGUID, triggerName, destGUID, destName, 0);
	end
	
	local effectEntry = activeAreaEffects[triggerGUID];			
	local destEffects = activeEffectsBySpell[destGUID];
	
	if(not destEffects) then
		-- Quality of 1.1 to enforce message
		destEffects = {[-1] = 0, [-2] = 1.1};
	
		activeEffectsBySpell[destGUID] = destEffects;
		activeEffectsByPriority[destGUID] = {};
		
	-- At the moment it is impossible for such an effect to be refreshed
	-- Since it is created by a summoned unit radiating it, it either
	-- gets removed/reapplied or removed/applied by a different unit.					
	elseif(destEffects[spellId]) then		
		error("Called ApplyAreaEffect on refreshed aura");
		return;
	end
	
	destEffects[spellId] = effectEntry;				
	effectEntry[9] = effectEntry[9] + 1; -- increase refcount
	
	-- While we keep the trigger itself as a normal afflicted unit to keep
	-- track when the area effect breaks, we do not broadcast it nor have to
	-- keep a sorted priority list
	if(triggerGUID ~= destGUID) then
		tinsert(activeEffectsByPriority[destGUID], effectEntry);
		sort(activeEffectsByPriority[destGUID], SortEffects);		
	
		-- Note that we CANNOT use nil as an amount, since external addons can rely on this value being non-nil
		-- for sorting. We're using -1 here that usually represents infinite values
		AM_Core:SendMessage("Absorbs_EffectApplied", triggerGUID, triggerName, destGUID, destName, spellId, -1, effectEntry[5], nil);
		
		-- Update quality if needed
		if(effectEntry[5] < destEffects[-2]) then
			destEffects[-2] = effectEntry[5];
		
			AM_Core:SendMessage("Absorbs_UnitUpdated", destGUID, destEffects[-1], destEffects[-2]);
		end
	end
	
	--AM_Core:Print("Applied area effect "..spellId.." (priority: "..effectData[1]..") on "..destGUID.." for "..value..", new total: "..activeEffectsBySpell[destGUID][0].." ("..#(activeEffectsByPriority[destGUID])..")");
end

local function CreateAreaTrigger(sourceGUID, sourceName, triggerGUID, triggerName, spellId)
	if(activeAreaEffects[triggerGUID]) then
		error("Trying to create new area trigger on existing one, triggerGUID: "..triggerGUID..", existing spellId: "..activeAreaEffects[triggerGUID][1]);
		return;
	end		

	local effectInfo = AbsorbEffects[spellId];
	
	local value, quality, extra = effectInfo[3](sourceGUID, sourceName, destGUID, destName, spellId, nil);
			
	if(value == nil) then return; end
	
	--AM_Core:Print("Creating area tigger "..triggerGUID.."/"..triggerName.." from "..sourceName.." with spellId "..spellId.." for "..value.."/"..quality);	
	
	local effectEntry = {spellId, -1 * effectInfo[1], value, value, quality, 0, extra, triggerGUID, 0};
	effectEntry[6] = AM_Core:ScheduleTimer(OnAreaTimeout, effectInfo[2] + 5, effectEntry);
	
	activeAreaEffects[triggerGUID] = effectEntry;
	
	AM_Core:SendMessage("Absorbs_AreaCreated", sourceGUID, sourceName, triggerGUID, spellId, value, quality);
end

-- Local variable to speed up access
local lowValueTolerance;
function AM_Core.SetLowValueTolerance(value)
	lowValueTolerance = value;
end

local AM_SendMessage = AM_Core.SendMessage;

local function HitUnit(guid, absorbedTotal, overkill, spellSchool)
	local guidEffects = activeEffectsBySpell[guid];

	if(not guidEffects) then return; end
	
	local absorbedRemaining = absorbedTotal;
	
	local absorbed = 0;
	local keepEffect, visibleAbsorb = false, false;
	local i = 1;
	local effectEntry, effectInfo;
	
	-- This loop lasts as long as there is still an absorb value that
	-- no effect could account for, but it will break once the list of
	-- available effects got used completely.
	while(absorbedRemaining > 0) do
		effectEntry = activeEffectsByPriority[guid][i];
		
		-- Sometimes there can be holes in this list since we don't re-sort
		-- after removing an effect
		if(effectEntry == nil) then break; end
		
		-- Only absorb effects with exactly zero are ignored, negative ones
		-- are treated as infinite (that is, no addon should display their value)
		if(effectEntry[3] ~= 0) then
			-- Hit the abosrb effect
			absorbed, keepEffect = AbsorbEffects[effectEntry[1]][4](effectEntry, absorbedRemaining, overkill, spellSchool);
			
			if(absorbed > 0) then
				-- Reduce the value of this effect and the remaining absorb value
				-- to be accounted for
				effectEntry[3] = effectEntry[3] - absorbed;
				absorbedRemaining = absorbedRemaining - absorbed;

				if(effectEntry[8]) then
					AM_SendMessage(AM_Core, "Absorbs_AreaUpdated", effectEntry[8], effectEntry[3]);
				
					--AM_Core:Print("Hit area effect #"..i..", absorbed "..absorbed.." (keep: "..tostring(keepEffect).."), remaining: "..absorbedRemaining.." - unit: "..effectEntry[3].."/"..activeEffectsBySpell[guid][0]);				
				else				
					-- If it should be visible (priority < 2), correct the total value
					if(effectEntry[2] < 2) then
						guidEffects[-1] = guidEffects[-1] - absorbed;
				
						-- Shows us that at least one visible effect got hit
						visibleAbsorb = true;
					end
					
					AM_SendMessage(AM_Core, "Absorbs_EffectUpdated", guid, effectEntry[1], effectEntry[3]);
					
					--AM_Core:Print("Hit effect #"..i..", absorbed "..absorbed.." (keep: "..tostring(keepEffect).."), remaining: "..absorbedRemaining.." - total: "..guidEffects[-1]);
				end						
			end
			
			-- If the hit-function told us to remove the effect, do so
			-- Note that only RemoveActiveEffect is allowed to remove it from the
			-- list, we just set the value to zero, so it gets ignored on any hit.
			-- This is do not come into any desync issues with the events, and to
			-- keep the proper clean-up code in one place
			if(not keepEffect) then				
				effectEntry[3] = 0;
			end
		end
			
		i = i + 1;
	end	
	
	-- there are two possibilities when things are going wrong
	-- 
	--	a)	we guessed an absorb value too high, in that case it will
	--		automatically be corrected when it breaks
	--
	--	b)	we guessed an absorb value too low, so we end up with an
	--		amount to absorb when all effects seem to be gone
	--		(absorbedRemaining > 0)
	--		Note that we cannot rely on SPELL_AURA_REMOVED to check this,
	--		since it may happen completely out of order, but it will
	--		clear this unit soon or did so already.
	--		we reduce the quality to zero, since any absorb now happening
	--		cannot be accounted for.
	--		Since we may have rounding errors from scanning the spellbook
	--		and calculating the value thereafter (does Blizzard round on
	--		EVERY step?!?), we accept a small threshold
	if(absorbedRemaining > lowValueTolerance) then
		guidEffects[-1] = guidEffects[-1] - absorbedRemaining;
		guidEffects[-2] = 0.0;
		visibleAbsorb = true;
	end
	
	if(visibleAbsorb) then
		AM_SendMessage(AM_Core, "Absorbs_UnitUpdated", guid, guidEffects[-1], guidEffects[-2]);
	end	
end

-- After profiling this method over several raiding nights, the custom triggers increased
-- the execution time (without its called functions) on average from 0.7 ys to 2.8 ys.
-- Given we have less than 100 combat log events happening per second, it's an acceptable
-- tradeoff given the features we get
local function OnCombatLogEvent(_, timestamp, type, sourceGUID, sourceName, sourceFlags, destGUID, destName, destFlags, ...)
	lastCombatLogEvent = timestamp;

	if(type == "SWING_DAMAGE") then
		local amount, overkill, school, resisted, blocked, absorbed = select(1, ...);
			
		if(not absorbed) then return; end
		
		HitUnit(destGUID, absorbed, amount, SCHOOL_MASK_PHYSICAL);
		
	elseif(type == "RANGE_DAMAGE" or type == "SPELL_DAMAGE" or type == "SPELL_PERIODIC_DAMAGE") then
		local spellId, spellName, spellSchool, amount, overkill, school, resisted, blocked, absorbed = select(1, ...);
		
		if(not absorbed) then return; end
		
		HitUnit(destGUID, absorbed, amount, spellSchool);		
		
	elseif(type == "SWING_MISSED") then
		local missType, amountMissed = select(1, ...);
		
		if(missType ~= "ABSORB") then return; end	
		
		HitUnit(destGUID, amountMissed, 0, SCHOOL_MASK_PHYSICAL);
		
	elseif(type == "RANGE_MISSED" or type == "SPELL_MISSED" or type == "SPELL_PERIODIC_MISSED") then
		local spellId, spellName, spellSchool, missType, amountMissed = select(1, ...);
			
		if(missType ~= "ABSORB") then return; end
		
		HitUnit(destGUID, amountMissed, 0, spellSchool);
		
	elseif(type == "SPELL_HEAL") then
		local spellId, spellName, spellSchool, amount, overhealing, absorb, critical = select(1, ...);		
		
		if(CombatTriggersOnHeal[sourceGUID]) then
			CombatTriggersOnHeal[sourceGUID](sourceGUID, sourceName, destGUID, destName, spellId, amount, overhealing);
		end
		
		if(critical and CombatTriggersOnHealCrit[sourceGUID]) then				
			CombatTriggersOnHealCrit[sourceGUID](sourceGUID, sourceName, destGUID, destName, spellId, amount, overhealing);
		end
	elseif(type == "SPELL_AURA_APPLIED") then
		local spellId, spellName = select(1, ...);
		
		if(AbsorbEffects[spellId]) then
			if(AbsorbEffects[spellId][1] > 0) then
				ApplySingularEffect(sourceGUID, sourceName, destGUID, destName, spellId);
			else
				ApplyAreaEffect(sourceGUID, sourceName, destGUID, destName, spellId);
			end
		end
		
		if(CombatTriggersOnAuraApplied[spellId]) then
			CombatTriggersOnAuraApplied[spellId](sourceGUID, sourceName, destGUID, destName, spellId);
		end
		
	elseif(type == "SPELL_AURA_REFRESH") then
		local spellId, spellName = select(1, ...);
		
		if(AbsorbEffects[spellId]) then
			if(AbsorbEffects[spellId][1] > 0) then
				ApplySingularEffect(sourceGUID, sourceName, destGUID, destName, spellId);
			else
				ApplyAreaEffect(sourceGUID, sourceName, destGUID, destName, spellId);
			end
		end
		
		if(CombatTriggersOnAuraApplied[spellId]) then
			CombatTriggersOnAuraApplied[spellId](sourceGUID, sourceName, destGUID, destName, spellId);
		end
				
	elseif(type == "SPELL_AURA_REMOVED") then
		local spellId = select(1, ...);			
		
		if(AbsorbEffects[spellId]) then
			if(activeEffectsBySpell[destGUID] and activeEffectsBySpell[destGUID][spellId]) then
				RemoveActiveEffect(destGUID, spellId);
			end
		end
		
		if(CombatTriggersOnAuraRemoved[spellId]) then
			CombatTriggersOnAuraRemoved[spellId](sourceGUID, sourceName, destGUID, destName, spellId);
		end
		
	elseif(type == "SPELL_SUMMON") then
		local spellId = select(1, ...);
		
		if(AreaTriggers[spellId]) then
			CreateAreaTrigger(sourceGUID, sourceName, destGUID, destName, AreaTriggers[spellId]);
		end
		
		--CombatTriggers(spellId.."_OnSummon"](sourceGUID, sourceName, destGUID, destName, spellId);
	end
end

local function OnPartyEvent()
	-- Note that the order here is VERY important
	if(UnitInBattleground("player")) then	
		curChatChannel = "BATTLEGROUND";
	elseif(UnitInRaid("player")) then
		curChatChannel = "RAID";
	elseif(GetNumPartyMembers() > 0) then
		curChatChannel = "PARTY";
	else
		curChatChannel = nil;
	end
end

-- Come on, this whole stuff is bullshit anyway
-- and will hopefully be removed in Cataclysm
local function OnZoneChangedNewArea()
	local profile = AM_Core.db.profile;
	
	if(not profile.zoneModifier_auto) then return; end

	if(UnitInBattleground("player")) then
		profile.zoneModifier = 1.17;
	elseif(IsActiveBattlefieldArena()) then
		profile.zoneModifier = 0.9;
	elseif(GetRealZoneText() == L["Wintergrasp"]) then
		-- 1.3 * 0.9 = fail Blizzard
		profile.zoneModifier = 1.17;
	elseif(GetRealZoneText() == L["Icecrown Citadel"]) then
		profile.zoneModifier = 1.3;
	else
		profile.zoneModifier = 1.3;
	end
	
	AM_Core.SetZoneModifier(profile.zoneModifier);
end

local lastAP, lastSP = 0, 0;

local function Comm_UnitStats()
	if(curChatChannel) then
		local curAP, curSP = UnitStats[playerGUID][2], UnitStats[playerGUID][3];
		
		if( (curAP ~= lastAP) or (curSP ~= lastSP) ) then
			AM_Core:SendCommMessage("Absorbs_UnitStats", AM_Core:Serialize(playerGUID, playerClass, curAP, curSP), curChatChannel);
			
			lastAP, lastSP = curAP, curSP;
			
			CommStatsCooldown = true;
			AM_Core:ScheduleTimer(ClearCommStatsCooldown, 15);
		end				
	end
end

local function Comm_Scaling()
	if(curChatChannel) then	
		AM_Core:SendCommMessage("Absorbs_Scaling", AM_Core:Serialize(playerGUID, playerClass, AM_Public.OnScalingEncode()), curChatChannel);
		
		CommScalingCooldown = true;
		AM_Core:ScheduleTimer(ClearCommScalingCooldown, 30);
	end
end

local function OnUnitStatsChanged(event, unit)
	local baseAP, plusAP, minusAP = UnitAttackPower("player");

	UnitStats[playerGUID][2] = baseAP + plusAP - minusAP;
	-- TODO: What about spell power ~= healing spell power?
	UnitStats[playerGUID][3] = GetSpellBonusHealing();

	if(curChatChannel) then
		AM_Core:ScheduleUniqueTimer("comm_stats", Comm_UnitStats, CommStatsCooldown and 15 or 5);
	end
end

local function OnUnitStatsReceived(prefix, text, distribution, target)
	if(not text) then return; end
	
	local success, guid, class, ap, sp = AM_Core:Deserialize(text);
	
	if(not(success and guid and class and ap and sp)) then return; end
	
	if(guid == playerGUID) then return; end
	
	if(not UnitStats[guid]) then
		UnitStats[guid] = {class, ap, sp, 1.0};
	else
		UnitStats[guid][2] = ap;
		UnitStats[guid][3] = sp;
		UnitStats[guid][4] = 1.0;
	end
end

local function OnScalingReceived(prefix, text, distribution, target)
	if(not text) then return; end
	
	local success, guid, class, inScaling = AM_Core:Deserialize(text);		
	
	if(not(success and guid and class and inScaling)) then return; end
	
	if(guid == playerGUID) then return; end
	
	(AM_Public.OnScalingDecode[class])(guid, inScaling);
end

local function OnPeriodicBroadcast()
	if(not curChatChannel) then return; end

	local playerStats = UnitStats[playerGUID];
	local playerScaling = Scaling[playerGUID];

	if(not CommStatsCooldown) then
		AM_Core:ScheduleUniqueTimer("comm_stats", Comm_UnitStats, 5);
	end
	
	if(not CommScalingCooldown) then
		AM_Core:ScheduleUniqueTimer("comm_scaling", Comm_Scaling, 5);
	end
end

local function OnInitialize()
	AM_Core:InitOptions();
	
	-- To actually see syntax errors in Effects.lua
	if(not AM_Public.CombatTriggers) then return; end
	
	--[[
	setmetatable(AM_Public.CombatTriggers, {
		__index = function()
			return DummyFunc;
		end
	});]]
	
	CombatTriggersOnHeal = AM_Public.CombatTriggers.OnHeal;
	CombatTriggersOnHealCrit = AM_Public.CombatTriggers.OnHealCrit;
	CombatTriggersOnAuraApplied = AM_Public.CombatTriggers.OnAuraApplied;
	CombatTriggersOnAuraRemoved = AM_Public.CombatTriggers.OnAuraRemoved;
	
	setmetatable(AM_Public.OnScalingDecode, {
		__index = function(table, class)
			return table.DEFAULT;
		end
	}); 
	
	AbsorbEffects = AM_Public.Effects;
	--CombatTriggers = AM_Public.CombatTriggers;
	AreaTriggers = AM_Public.AreaTriggers;
end

local function OnEnable()
	_, playerClass = UnitClass("player");
	playerGUID = UnitGUID("player");
	
	UnitStats = {[playerGUID] = {playerClass, 0, 0, 1.0}};
	Scaling = {[-1] = {}, [playerGUID] = {}}
	
	AM_Core:RegisterEvent("ZONE_CHANGED_NEW_AREA", OnZoneChangedNewArea);
	OnZoneChangedNewArea();

	if(playerClass == "DEATHKNIGHT") then
		AM_Core:RegisterEvent("UNIT_ATTACK_POWER", OnUnitStatsChanged);
		
	elseif(playerClass == "DRUID") then
		AM_Core:RegisterEvent("UNIT_ATTACK_POWER", OnUnitStatsChanged);
		
	elseif(playerClass == "MAGE") then
		AM_Core:RegisterEvent("PLAYER_DAMAGE_DONE_MODS", OnUnitStatsChanged);
		
	elseif(playerClass == "PALADIN") then
		AM_Core:RegisterEvent("PLAYER_DAMAGE_DONE_MODS", OnUnitStatsChanged);
		
	elseif(playerClass == "PRIEST") then
		AM_Core:RegisterEvent("PLAYER_DAMAGE_DONE_MODS", OnUnitStatsChanged);
		
	elseif(playerClass == "WARLOCK") then
		AM_Core:RegisterEvent("PLAYER_DAMAGE_DONE_MODS", OnUnitStatsChanged);
	end
	
	OnUnitStatsChanged();
			
	-- This has to happen before class init, else we get no initial scaling broadcast
	if(not AM_Core.db.profile.silent) then
		AM_Core:SetVerbose();
	end
			
	AM_Core:EnableEffects();
	
	if(AM_Public.OnEnableClass[playerClass]) then
		AM_Public.OnEnableClass[playerClass]();
	end
	
	if(AM_Public.OnLevelUp) then
		AM_Core:RegisterEvent("PLAYER_LEVEL_UP", AM_Public.OnLevelUp);
	end
	
	if(AM_Public.OnTalentUpdate) then
		AM_Core:RegisterEvent("PLAYER_TALENT_UPDATE", AM_Public.OnTalentUpdate);
	end
	
	if(AM_Public.OnGlyphUpdated) then
		AM_Core:RegisterEvent("GLYPH_UPDATED", AM_Public.OnGlyphUpdated);
	end
	
	if(AM_Public.OnEquipmentChanged) then
		AM_Core:RegisterEvent("PLAYER_EQUIPMENT_CHANGED", AM_Public.OnEquipmentChanged);
	end
	
	AM_Core:ScheduleRepeatingTimer(OnPeriodicBroadcast, 300);
	
	AM_Core:RegisterComm("Absorbs_UnitStats", OnUnitStatsReceived);
	AM_Core:RegisterComm("Absorbs_Scaling", OnScalingReceived);	
	
	local profile = AM_Core.db.profile;
	
	AM_Core.SetLowValueTolerance(profile.lowValueTolerance);	
	AM_Core.SetZoneModifier(profile.zoneModifier);	
		
	if(not profile.passive) then
		AM_Core:SetActive();
	end	
end

local function OnDisable()
	-- Since we cannot make any assumption about the data when
	-- we come back, screw it all and reinitialize
	
	for guid, effects in pairs(activeEffectsBySpell) do
		AM_Core:SendMessage("Absorbs_UnitCleared", guid);
	end
	
	activeEffectsBySpell = {};
	activeEffectsByPriority = {};
	activeAreaEffects = {};	
	activeCharges = {};
	
	Scaling = nil;
	UnitStats = nil;
	
	AM_Core:UnregisterAllEvents();
	AM_Core:UnregisterAllMessages();
	AM_Core:CancelAllTimers();
	
	collectgarbage("collect");
end


-- #################################################################################################


-- An extension of AceTimer to schedule a one-shot timer that will not
-- be scheduled twice if scheduled again before it's fired.
local activeTimers = {};

function AM_Core:ScheduleUniqueTimer(id, callback, delay, arg)
	if(not activeTimers[id]) then
		self:ScheduleTimer(
			function()
				callback(arg);
				activeTimers[id] = nil;
			end, delay);
			
		activeTimers[id] = 1;
	end
end

function AM_Core:SetActive()
	self:RegisterEvent("COMBAT_LOG_EVENT_UNFILTERED", OnCombatLogEvent);	
	self.db.profile.passive = false;
end

function AM_Core:SetPassive()
	self:UnregisterEvent("COMBAT_LOG_EVENT_UNFILTERED");
	self.db.profile.passive = true;
end

function AM_Core:SetVerbose()
	self:RegisterEvent("PARTY_MEMBERS_CHANGED", OnPartyEvent);
	self:RegisterEvent("RAID_ROSTER_UPDATE", OnPartyEvent);
	self.db.profile.silent = false;
	
	OnPartyEvent();
end

function AM_Core:SetSilent()
	self:UnregisterEvent("PARTY_MEMBERS_CHANGED");
	self:UnregisterEvent("RAID_ROSTER_UPDATE");
	self.db.profile.silent = true;
	
	curChatChannel = nil;
end


-- #################################################################################################


AM_Public.OnInitialize = OnInitialize;

AM_Public.OnEnable = function()
	if(not GetTalentInfo(1, 1)) then
		AM_Core:RegisterEvent("PLAYER_ALIVE",
			function()
				OnEnable();
				AM_Core:UnregisterEvent("PLAYER_ALIVE");
			end
		);
	else
		OnEnable();
	end	
end

AM_Public.OnDisable = OnDisable;

function AM_Public.GetCoreObject()
	return AM_Core;
end

function AM_Public.OnScalingEncode()
	return Scaling[playerGUID];
end

AM_Public.OnScalingDecode.DEFAULT = function(guid, guidScaling)
	Scaling[guid] = guidScaling;
end

function AM_Public.PrintProfiling()
	if(GetCVar("scriptProfile") ~= "1") then
		AM_Core:Print("CPU profiling disabled");
		return;
	end

	UpdateAddOnCPUUsage();
	UpdateAddOnMemoryUsage();
	
	AM_Core:Print("Mem: "..format("%.3f", GetAddOnMemoryUsage("AbsorbsMonitor")).." kB");
	AM_Core:Print("Time: "..format("%.3f", GetAddOnCPUUsage("AbsorbsMonitor")).." ms");
	AM_Core:Print("--- critical code paths (all times in ms) ---");
	
	local funcTable;
	
	funcTable = {
		["ApplySingularEffect"] = ApplySingularEffect,
		["HitUnit"] = HitUnit,
		["RemoveActiveEffect"] = RemoveActiveEffect,
		["OnCombatLogEvent"] = OnCombatLogEvent,
		["SortEffects"] = SortEffects,
	};
		
	local v_type;
	local time_self, time_combined, count;
	
	for k, v in pairs(funcTable) do
		v_type = type(v);
		
		if(v_type == "function") then
			time_self, count = GetFunctionCPUUsage(v, false);
			time_combined = GetFunctionCPUUsage(v, true);
			
			AM_Core:Print(k.." (#"..count.."): "..format("%.4f", time_self).." / "..format("%.4f", time_combined));
		end
	end
end

-- This is uses a _very_ simple queue implementation with tinsert and tremove.
-- It will not scale very well for large values, but we're talking of a maximum
-- of ~3 entries per GUID at any given time. A proper implementation
-- with linked list would probably not be any faster.
function AM_Public.PushCharge(guid, spellId, amount, lifetime)
	local guidCharges = activeCharges[guid];

	if(not guidCharges) then
		activeCharges[guid] = { [spellId] = {lastCombatLogEvent + lifetime, amount} };
		
	elseif(not guidCharges[spellId]) then
		guidCharges[spellId] = {lastCombatLogEvent + lifetime, amount};
	
	else
		tinsert(guidCharges[spellId], lastCombatLogEvent + lifetime);
		tinsert(guidCharges[spellId], amount);
	end
end

function AM_Public.PopCharge(guid, spellId)
	local guidCharges = activeCharges[guid];
	
	if(guidCharges) then
		local queue = activeCharges[guid][spellId];
		
		-- For some weird reason, it will fail (true even on empty array)
		-- if checked for queue[1] ?!?
		if(queue and queue[2]) then
			local chargeAmount;
			local chargeExpire;			
		
			-- This loop will not be able to run infinitely
			while(true) do
				-- In this order we might save one table reshuffle
				chargeAmount = tremove(queue, 2);
				
				if(not chargeAmount) then return 0; end
				
				chargeExpire = tremove(queue, 1);
			
				if(chargeExpire > lastCombatLogEvent) then
					return chargeAmount;
				end
			end
		end
	end
	
	return 0;
end

function AM_Public.AddCombatTrigger(target, event, func)
	local eventTriggers = AM_Public.CombatTriggers[event];
	local oldTrigger = eventTriggers[target];

	if(not oldTrigger) then
		eventTriggers[target] = func;
		
	else
		local listIndex = target.."_list";
		local funcList = eventTriggers[listIndex];
		
		-- There is already a list of callbacks, so we just add this one
		if(funcList) then
			for k, v in pairs(funcList) do
				if(v == func) then
					return;
				end
			end
		
			tinsert(funcList, func);
			
		-- We used a direct call so far, create a list and set up a handler
		else
			if(oldTrigger == func) then
				return;
			end
		
			funcList = {oldTrigger, func};			
			eventTriggers[listIndex] = funcList;
			
			local handler = function(...)
				for k, v in pairs(funcList) do
					v(...);
				end
			end
			
			eventTriggers[target] = handler;
		end		
	end
end

function AM_Public.RemoveCombatTrigger(target, event, func)
	local eventTriggers = AM_Public.CombatTriggers[event];
	
	local listIndex = target.."_list";
	local funcList = eventTriggers[listIndex];
	
	-- We have a list of callbacks, reduce if possible
	if(funcList) then
		if(#funcList == 2) then
			eventTriggers[target] = (funcList[1] == func) and funcList[2] or funcList[1];
			eventTriggers[listIndex] = nil;
		else
			-- ATTENTION: We have to keep the table in place
			-- because the handler references this table
			
			local old_funcList = DeepTableCopy(funcList);
			wipe(funcList);
			
			for k, v in pairs(old_funcList) do
				if(v ~= func) then				
					tinsert(funcList, v);
				end
			end
		end
	
	-- It was a direct call anyway
	else
		eventTriggers[target] = nil;
	end	
end

function AM_Public.Unit_Total(guid)
	local guidEffects = activeEffectsBySpell[guid];
	
	return (guidEffects and guidEffects[-1] or 0);
end

function AM_Public.Unit_Effect(guid, spellId)
	local guidEffects = activeEffectsBySpell[guid];
	
	if(guidEffects) then
		if(guidEffects[spellId]) then
			return guidEffects[spellId][3];
		end
	end
	
	return 0;
end

function AM_Public.Unit_Stats(guid, missingQuality)
	local guidStats = UnitStats[guid];
	
	if(guidStats) then
		return guidStats[2], guidStats[3], guidStats[4];
	else
		return 0, 0, missingQuality;
	end		
end

function AM_Public.Unit_ScalingAll()
	return Scaling;
end

function AM_Public.Unit_Scaling(guid, defaultScaling, defaultQuality)
	local guidScaling = Scaling[guid];

	if(guidScaling) then
		return guidScaling, 1.0;
	else
		return defaultScaling, defaultQuality;
	end
end

-- Optimized method to save one function call on creation, since a lot of spells
-- actually require stats and scaling
function AM_Public.Unit_StatsAndScaling(guid, missingQuality, defaultScaling, defaultQuality)
	local guidStats = UnitStats[guid];
	local guidScaling = Scaling[guid];
	
	if(guidStats) then
		if(guidScaling) then
			return guidStats[2], guidStats[3], guidStats[4], guidScaling, 1.0;
		else
			return guidStats[2], guidStats[3], guidStats[4], defaultScaling, defaultQuality;
		end
	else
		if(guidScaling) then
			return 0, 0, missingQuality, guidScaling, 1.0;
		else
			return 0, 0, missingQuality, defaultScaling, defaultQuality;
		end
	end
end

function AM_Public.Unit_EffectsMap(guid)
	return activeEffectsBySpell[guid];
end

function AM_Public.Unit_EffectsList(guid)
	return activeEffectsByPriority[guid];
end

function AM_Public.ScheduleScalingBroadcast()
	if(curChatChannel) then
		AM_Core:ScheduleUniqueTimer("comm_scaling", Comm_Scaling, CommScalingCooldown and 30 or 5);
	end
end

function AM_Public.Test()	
	ApplySingularEffect("0x0", "source", "0x1a", "dest A", 27779);
	ApplySingularEffect("0x0", "source", "0x1a", "dest A", 48066);
	ApplySingularEffect("0x0", "source", "0x1b", "dest B", 31771);
	ApplySingularEffect("0x0", "source", "0x1b", "dest B", 57350);
	ApplySingularEffect("0x0", "source", "0x1c", "dest C", 48066);
	ApplySingularEffect("0x0", "source", "0x1c", "dest C", 28810);
end

-- Absorbs_EffectApplied
-- The effect-individual messages get sent on visible and non-visible effects
-- sourceGUID, sourceName, destGUID, destName, spellId, value, quality, duration

-- Absorbs_EffectUpdated
-- guid, spellId, value, duration (only if refreshed)

-- Absorbs_EffectRemoved
-- guid, spellId

-- Whenever the unit that radiates an AREA effect is created (visible/non-visible)
-- Note that the actual effect on the unit that absorbs damage casues an
-- EffectApplied/EffectRemoved message, but not EffectUpdated (instead AreaUpdated)
-- The rationale behind this is performance, since we cannot update every unit afflicted
-- by the area effect on every hit. Therefore, we have the shared entry in the activeEffects
-- table of each unit, and will handle it the same way when exporting - separately from each
-- others.

-- Absorbs_AreaCreated
-- sourceGUID, sourceName, triggerGUID, spellId, value, quality

-- Absorbs_AreaUpdated
-- triggerGUID, value

-- Absorbs_AreaCleared
-- triggerGUID

-- Absorbs_UnitUpdated
-- Only for VISIBLE changes on the total amount
-- guid, value, quality

-- Absorbs_UnitCleared
-- Everytime a unit gets cleared from all absorb effects (quality reset)
-- including non-visible effects
-- guid
